"use server";

import Dockerode from "dockerode";
import { pullImage } from "@/actions/docker/image";
import {
  MessageType,
  MessageProps,
} from "@/app/playground/layout/terminal/components/message";

enum ResultType {
  CompilationError = "compilation error",
  RuntimeError = "runtime error",
  Success = "success",
}

interface RunCodeResult {
  success: boolean;
  output: string;
  message: string;
  type: ResultType;
}

async function prepareEnvironment(docker: Dockerode, imageGCC: string) {
  const images = await docker.listImages();
  const imageExists = images.some(
    (image) => image.RepoTags && image.RepoTags.includes(imageGCC)
  );

  if (!imageExists) {
    await pullImage(imageGCC);
  }
}

async function compileAndRun(
  code: string,
  language: string,
  docker: Dockerode,
  imageGCC: string
): Promise<Dockerode.Container> {
  let compileCommand = "";

  if (language === "c") {
    compileCommand = `gcc main.c -o main && ./main`;
  } else if (language === "cpp") {
    compileCommand = `g++ main.cpp -o main && ./main`;
  } else {
    throw new Error("Unsupported language: " + language);
  }

  const extension = language === "c" ? "c" : "cpp";

  const container = await docker.createContainer({
    Image: imageGCC,
    Cmd: [
      "sh",
      "-c",
      `export LANG=C.UTF-8 && printf '%s' '${code.replace(
        /'/g,
        "'\\''"
      )}' > main.${extension} && ${compileCommand}`,
    ],
    Tty: false,
  });

  await container.start();
  return container;
}

async function getOutputAndResultType(
  stream: NodeJS.ReadableStream
): Promise<RunCodeResult> {
  return new Promise((resolve, reject) => {
    let output = "";

    stream.on("data", (chunk) => {
      console.debug("Chunk received:", Buffer.from(chunk).toString("hex"));
      output += chunk.toString("utf8");
    });

    stream.on("end", () => {
      console.debug("Final output:", output);

      const isSuccessful = !output.includes("error:");
      const message = isSuccessful
        ? "运行成功：编译成功"
        : "运行错误：编译失败";

      resolve({
        success: isSuccessful,
        output: output.replace(/[^\x20-\x7E\n\r]/g, ""), // 移除非打印字符
        message,
        type: isSuccessful ? ResultType.Success : ResultType.CompilationError,
      });
    });

    stream.on("error", (err) => {
      reject(err);
    });
  });
}

async function cleanupContainer(container: Dockerode.Container): Promise<void> {
  try {
    const containerInfo = await container.inspect();
    if (containerInfo.State.Running) {
      await container.stop();
    }
    await container.remove();
  } catch (cleanupError) {
    console.error("清理容器时出错:", cleanupError);
  }
}

function mapResultTypeToMessageType(type: ResultType): MessageType {
  switch (type) {
    case ResultType.CompilationError:
    case ResultType.RuntimeError:
      return "error";
    case ResultType.Success:
      return "success";
    default:
      return "info";
  }
}

export async function runCode(
  language: string,
  code: string
): Promise<MessageProps & { output: string }> {
  const docker = new Dockerode({ socketPath: "/var/run/docker.sock" });
  const imageGCC = "gcc:latest";
  let container: Dockerode.Container | undefined;

  try {
    if (language === "c" || language === "cpp") {
      await prepareEnvironment(docker, imageGCC);
      container = await compileAndRun(code, language, docker, imageGCC);
      const stream = await container.attach({
        stream: true,
        stdout: true,
        stderr: true,
      });

      const result = await getOutputAndResultType(stream);
      return {
        type: mapResultTypeToMessageType(result.type),
        message: result.message,
        output: result.output,
      };
    } else {
      throw new Error(`Unsupported language: ${language}`);
    }
  } catch (error) {
    console.error("Error running code:", error);
    return {
      type: "error",
      message: error instanceof Error ? error.message : "Unknown error",
      output: "",
    };
  } finally {
    if (container) {
      await cleanupContainer(container);
    }
  }
}
